import typer
import importlib.metadata

from candie import parse, setup, execute

app = typer.Typer()


@app.command()
def x(
    s: str = typer.Argument(None, help="Execute a section."),
    v: bool = typer.Option(False, "--version", "-v", help="Show the version", is_eager=True)
) -> None:
    
    if v:
        typer.echo(f'''
candie (candie.kit) v{importlib.metadata.version('candie.kit')}
Copyright (C) 2025 Yash Tajne
This program may be freely redistributed under
the terms of the Apache Software License.
''')
        exit()    
    
    if s:
        sections = parse()
        setup(sections)
        execute(sections, s.upper())

def start() -> None:
    app()


import os
import io
import re
import sys
import shutil
import json
import zipfile
import requests
import pkgconfig
import subprocess


from . import C, PROJECT, LOCAL_INSTALL_DIR, PROJECT_ROOT, NATIVE, DOWNLOADS_DIR, _GLOBAL_NAMESPACE_, parse, setup, execute, Target
from .printer import Print_Error, console


# Download a project from remote and builds it
def Fetch_Content(
        url: str,
        place: str = DOWNLOADS_DIR
    ):

    console.print(f'[ Downloading ] {url}')
    os.makedirs(os.path.join(place, "dist"), exist_ok=True)

    with requests.get(url, stream=True) as response:
        if response.status_code == 200:
            content_disposition = response.headers.get("Content-Disposition")
            if content_disposition and "filename=" in content_disposition:
                zip_filename = content_disposition.split("filename=")[-1].strip('"')
            else:
                zip_filename = url.split("/")[-1]
            zip_path = os.path.join(place, "dist", zip_filename)

            # Writing stream
            with open(zip_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    f.write(chunk)

            #Unarchive zip file
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(place)

            # os.remove(zip_path)

            # Build file path
            buildfile = os.path.join((content_root := os.path.join(place, os.path.splitext(zip_filename)[0])), 'Candiefile')
            
            console.print("[ Switching ] PROJECT_ROOT")
            console.print("[ Executing ] BUILD section")

            _GLOBAL_NAMESPACE_['PROJECT_ROOT'] = content_root

            sections = parse(buildfile)
            setup(sections)
            execute(sections, 'BUILD')

            console.print("[ Reverting ] PROJECT_ROOT")

            _GLOBAL_NAMESPACE_['PROJECT_ROOT'] = PROJECT_ROOT

        else:
            Print_Error(f"failed to download. HTTP Response code {response.status_code}")


def Run_Command(
        *cmd,
        directory: str = PROJECT_ROOT, 
        shell: bool = False, 
        capture_output: bool = False
    ) -> subprocess.CompletedProcess:
    try:
        result = subprocess.run(
            list(cmd),
            cwd=directory,
            shell=shell,
            capture_output=capture_output,
            text=True,
            check=True
        )
        return result
    except subprocess.CalledProcessError as e:
        Print_Error(f"Command failed with return code {e.returncode} {e.cmd}")
        raise SystemError


# Finds the pkgconfig file of the library and returns its values.
def Grab_Dependency(
        name: str,
        debug: bool = False,
        target: Target = NATIVE,
        directory: str = None
    ) -> (dict | None):
    directory = os.path.join(LOCAL_INSTALL_DIR, target) if directory is None else directory
    os.environ['PKG_CONFIG_PATH'] = f'{directory}{'/debug' if debug else ''}/lib/pkgconfig'
    return {
        'name': name,
        'debug': debug,
        'cflags': pkgconfig.cflags(name),
        'libs': pkgconfig.libs(name)
    } if pkgconfig.exists(name) else None

# Returns a library name based on the type of library and operating system.
def Generate_Library_Name(library_name: str, library_type: str, os_triplet: str) -> str:
    os_triplet = os_triplet.lower()
    if "darwin" in os_triplet or "macos" in os_triplet:
        os_prefix = 'lib'
        if library_type == 'static':
            os_suffix = '.a'
        elif library_type == 'dynamic':
            os_suffix = '.dylib'
        else:
            raise ValueError('Unsupported library type')
    elif "windows" in os_triplet:
        os_prefix = ''
        if library_type == 'static':
            os_suffix = '.lib'
        elif library_type == 'dynamic':
            os_suffix = '.dll'
        else:
            raise ValueError('Unsupported library type')
    else:
        os_prefix = 'lib'
        if library_type == 'static':
            os_suffix = '.a'
        elif library_type == 'dynamic':
            os_suffix = '.so'
        else:
            raise ValueError('Unsupported library type')

    return f'{os_prefix}{library_name}{os_suffix}'

# Grabs the files files with the provided extension from a directory
def Grab_Files(directory_path: str, file_ext: str) -> tuple[str, ...]:  
    source_directory_path = os.path.abspath(os.path.join(_GLOBAL_NAMESPACE_.get('PROJECT_ROOT', PROJECT_ROOT), directory_path))
    return tuple(
        str(file.path) for file in os.scandir(source_directory_path)
        if file.is_file() and file.name.endswith(file_ext)
    )

# Grabs the source (.c , .cpp) files from the specified directory
def Grab_Sources(source_directory_path: str) -> tuple[str, ...]:
    if PROJECT['language'] == C:
        return Grab_Files(source_directory_path, '.c')
    else:
        return Grab_Files(source_directory_path, '.c') + Grab_Files(source_directory_path, '.cpp')

# Grabs the header (.h, .hpp) files from the specified directory
def Grab_Headers(source_directory_path: str) -> tuple[str, ...]:
    if PROJECT['language'] == C:
        return Grab_Files(source_directory_path, '.h')
    else:
        return Grab_Files(source_directory_path, '.h') + Grab_Files(source_directory_path, '.hpp')

# Copies directory files from source to destination direcory
def Copy_Directory(source, destination) -> list[str]:
    copied: list[str] = []
    if not os.path.exists(destination):
        os.makedirs(destination)
    for root, _, files in os.walk(source):
        rel_path = os.path.relpath(root, source)
        dest_subdir = os.path.join(destination, rel_path)
        if not os.path.exists(dest_subdir):
            os.makedirs(dest_subdir)
            copied.append(dest_subdir)
        for file in files:
            src_file = os.path.join(root, file)
            dest_file = os.path.join(dest_subdir, file)
            shutil.copy2(src_file, dest_file)
            copied.append(dest_file)
    return copied

# Copy pastes files.
def Copy_Files(files: list[str], directory: str):
    if not os.path.exists(directory):
        os.makedirs(directory)
    for file_path in files:
        file_name = os.path.basename(file_path)
        destination_file_path = os.path.join(directory, file_name)
        try:
            shutil.copy(file_path, destination_file_path)
            console.print(f"[ copied ] <{file_path}>")
        except Exception as e:
            Print_Error(f"failed to copy <{file_path}> -> {str(e)}")

def Generate_PkgConfig_File(
        name: str,
        description: str,
        version: str,
        url: str,
        requires: list[str],
        libs: list[str],
        cflags: list[str],
        debug: bool = False
    ):
    return f'''
prefix=${{pcfiledir}}/../..{'/..' if debug else ''}
exec_prefix=${{prefix}}
includedir=${{prefix}}/include
libdir=${{prefix}}/{'debug/lib' if debug else 'lib'}

Name: {name}
Description: {description}
Version: {version}
URL: {url}

Libs: "-L${{libdir}}" {' '.join(libs)}
Requires: {' '.join(requires)}
Cflags: "-I${{includedir}}" {' '.join(cflags)}
'''


import re
import os
import platform


class Arg(str):
    def __new__(cls, arg):
        obj = super().__new__(cls, arg)
        obj.value = arg
        return obj
    
class Target(str):
    def __new__(cls, cpu, os="", abi=""):
        if os and abi:
            triplet = f"{cpu}-{os}-{abi}"
        elif os:
            triplet = f"{cpu}-{os}"
        else:
            triplet = cpu
        obj = super().__new__(cls, triplet)
        return obj

_GLOBAL_NAMESPACE_ = globals()

PROJECT_ROOT = os.getcwd()
ARCHITECTURE = platform.architecture()
SYSTEM = platform.system()

# Files
CANDIEFILE = f"{PROJECT_ROOT}/Candiefile"

# Directories
CANDIE_DIR = f"{PROJECT_ROOT}/.candie"
LOCAL_INSTALL_DIR = f"{CANDIE_DIR}/installed"
DOWNLOADS_DIR = f"{CANDIE_DIR}/downloads"
CACHE_DIR = f"{CANDIE_DIR}/cache"

# Project object
PROJECT: dict = {}

# Languages
C:      str = '.c'
CPP:    str = '.cpp'

# Library types
STATIC = 'static'
SHARED = 'dynamic'

# Triplets
TARGETS = [
    (NATIVE := Target('native')),

    (X86_WINDOWS    := Target('x86',     'windows')),
    (X64_WINDOWS    := Target('x86_64',  'windows')),
    (ARM64_WINDOWS  := Target('aarch64', 'windows')),

    (X86_LINUX      := Target('x86',     'linux')),
    (X64_LINUX      := Target('x86_64',  'linux')),
    (ARM64_LINUX    := Target('aarch64', 'linux')),

    (ARM64_MACOS    := Target('aarch64', 'macos')),
    (X86_MACOS      := Target('x86',     'macos')),
    (X64_MACOS      := Target('x86_64',  'macos'))
]

# Compiler Args
C_ARGS = [

    # Optimization
    (O0 := Arg('-O0')),
    (O1 := Arg('-O1')),
    (O2 := Arg('-O2')),
    (O3 := Arg('-O3')),
    (OS := Arg('-Os')),
    (OZ := Arg('-Oz')),

    # Warning and error
    (W_ALL      := Arg('-Wall')),
    (W_EXTRA    := Arg('-Wextra')),
    (W_ERROR    := Arg('-Werror')),
    (W_PEDANTIC := Arg('-Wpedantic')),

    # Standard version c
    (STD_C89     := Arg('-std=c89')),
    (STD_C90     := Arg('-std=c90')),
    (STD_GNU89   := Arg('-std=gnu89')),
    (STD_C99     := Arg('-std=c99')),
    (STD_GNU99   := Arg('-std=gnu99')),
    (STD_C11     := Arg('-std=c11')),
    (STD_GNU11   := Arg('-std=gnu11')),
    (STD_C17     := Arg('-std=c17')),
    (STD_GNU17   := Arg('-std=gnu17')),
    (STD_C23     := Arg('-std=c23')),
    (STD_GNU23   := Arg('-std=gnu23')),

    # Standard version c++
    (STD_CPP98   := Arg('-std=c++98')),
    (STD_GNUPP98 := Arg('-std=gnu++98')),
    (STD_CPP03   := Arg('-std=c++03')),
    (STD_GNUPP03 := Arg('-std=gnu++03')),
    (STD_CPP11   := Arg('-std=c++11')),
    (STD_GNUPP11 := Arg('-std=gnu++11')),
    (STD_CPP014  := Arg('-std=c++14')),
    (STD_GNUPP14 := Arg('-std=gnu++14')),
    (STD_CPP17   := Arg('-std=c++17')),
    (STD_GNUPP17 := Arg('-std=gnu++17')),
    (STD_CPP20   := Arg('-std=c++20')),
    (STD_GNUPP20 := Arg('-std=gnu++20')),
    (STD_CPP23   := Arg('-std=c++23')),
    (STD_GNUPP23 := Arg('-std=gnu++23')),
]

# Linker arguments
LINK_ARGS = [

    # Optimiztion & Size reduction
    (WL_O1          := Arg('-Wl,-O1')),
    (WL_O2          := Arg('-Wl,-O2')),
    (WL_ASNEEDED    := Arg('-Wl,--as-needed')),
    (WL_GC_SECTIONS := Arg('-Wl,--gc-sections')),
    (WL_STRIP_ALL   := Arg('-Wl,--strip-all')),
    (WL_STRIP_DEBUG := Arg('-Wl,--strip-debug')),
    (WL_ICF_SAFE    := Arg('-Wl,--icf=safe')),

    # For Performance
    (WL_HASH_STYLE_GNU := Arg('-Wl,--hash-style=gnu')),
    (WL_NO_KEEP_MEMORY := Arg('-Wl,--no-keep-memory')),
    (WL_THREADS        := Arg('-Wl,--threads')),

    # Debugging & Profiling
    (WL_EXPORT_DYNAMIC := Arg('-Wl,--export-dynamic')),
    (WL_PRINT_MAP      := Arg('-Wl,--print-map')),
    (WL_VERBOSE        := Arg('-Wl,--verbose')),
    (WL_WARN_COMMON    := Arg('-Wl,--warn-common')),
]


def parse(file: str = CANDIEFILE) -> dict:
    if not os.path.isfile(file):
        Print_Error("Build instruction file not found.")
        raise SystemExit

    with open(file, 'r') as f:
        data = f.read()
    sections = re.findall(r'^\[(.*?)\](.*?)(?=^\[|\Z)', data, re.DOTALL | re.MULTILINE)
    parsed_data = {section.strip(): content.strip() for section, content in sections}
    return parsed_data

def setup(sections: dict) -> None:
    
    for dir in [
        CANDIE_DIR,
        CACHE_DIR,
        LOCAL_INSTALL_DIR,
        DOWNLOADS_DIR
    ]:
        if not os.path.isdir(dir):
            os.makedirs(dir)
    
    exec(sections["SETUP"], _GLOBAL_NAMESPACE_)
    
    # console.print("Sections:", sections)
    # console.print("Project:", PROJECT)
    
def execute(sections: dict, option: str) -> None:
    exec(sections[option], _GLOBAL_NAMESPACE_)


from .utils import (
    Grab_Files,
    Grab_Sources,
    Grab_Headers,
    Grab_Dependency,

    Copy_Directory,
    Copy_Files,

    Fetch_Content,
    Run_Command,
)

from .genrators import Executable, Package, Library
from .compiler import ZigToolchain as Compiler
from .printer import Print_Error

grab_files       = Grab_Files
grab_sources     = Grab_Sources
grab_headers     = Grab_Headers
grab_dependency  = Grab_Dependency

copy_files       = Copy_Files
copy_directory   = Copy_Directory

fetch_content    = Fetch_Content
run_command      = Run_Command

executable       = Executable
library          = Library
package          = Package

def project(name: str, language: str, version: str = '1.0.0') -> None:
    PROJECT['name'] = name
    PROJECT['language'] = language
    PROJECT['version'] = version

    if language not in [C, CPP]:
        Print_Error("Invalid Language", language)

    PROJECT['compiler'] = Compiler(PROJECT['language'], CACHE_DIR)




from rich.console import Console


console = Console()


def Print_Error(err: str):
    console.print('[ [bright_red]Error[/bright_red] ] ', err)

def Print_Warning(warn: str):
    console.print('[ [deep_pink2]Warning[/deep_pink2] ] ', warn)

def Print_Msg(label, msg: str, end: str = '\n'):
    console.print(f'[ [chartreuse1]{label}[/chartreuse1] ] ', msg, end=end)